-- packages/generic_trees/sql/generic-trees-package.sql
--
-- @creation-date December 2000
--
-- @author Aram Kananov (akananov@arsdigita.com)
--
-- @cvs-id $Id: generic-trees-package.sql,v 1.1 2001/04/27 10:03:33 arcanan Exp $
--
-- TODO LIST:
-- 1)index api
-- 2)check drop objects

create or replace package generic_tree_type  
as 
    FUNCTION new ( 
        type_id         in generic_tree_types.type_id%TYPE	    default null, 
	owner_id            in generic_tree_types.owner_id%TYPE         default null, 
	type_name	    in generic_tree_types.type_name%TYPE, 
	pretty_name	    in generic_tree_types.pretty_name%TYPE, 
	pretty_plural	    in generic_tree_types.pretty_plural%TYPE, 
	tree_struct_table   in generic_tree_types.tree_struct_table%TYPE, 
	default_function_no in generic_tree_types.default_function_no%TYPE default null,
	creation_date       in acs_objects.creation_date%TYPE		    default sysdate, 
	creation_user       in acs_objects.creation_user%TYPE		    default null, 
	creation_ip         in acs_objects.creation_ip%TYPE		    default null, 
	context_id          in acs_objects.context_id%TYPE		    default null
    ) RETURN generic_tree_types.type_id%TYPE
    ;

    FUNCTION name (
        object_id           in acs_objects.object_id%TYPE
    ) return varchar2;

    PROCEDURE delete ( 
	type_id         in acs_objects.object_id%TYPE 
    );

    FUNCTION check_before_create (
	p_type_id       generic_tree_types.type_id%TYPE,
	error_descr	    in out varchar2
    ) RETURN integer;

    FUNCTION generate_db_objects (
	p_type_id       generic_tree_types.type_id%TYPE,
	error_descr	    in out varchar2
    ) RETURN integer;

    FUNCTION drop_db_objects (
        p_type_id       generic_tree_types.type_id%TYPE,
	errors_descr in out varchar2
    ) RETURN integer;
end; 
/ 
show errors 

CREATE OR REPLACE PACKAGE BODY generic_tree_type AS
    ------------------------------------------------------------
    -- LOCAL FUNCTIONS and PROCEDURES DEFINITIONS
    ------------------------------------------------------------

    FUNCTION check_before_create (
	p_type_id       generic_tree_types.type_id%TYPE,
	error_descr	    in out varchar2
    ) RETURN integer
    IS
        is_any integer;
	trouble boolean := false;
	p_tree_struct_table generic_tree_types.tree_struct_table%TYPE;
    BEGIN
	select tree_struct_table 
	    into  p_tree_struct_table  
	    from generic_tree_types 
	    where type_id = p_type_id
	;
	select count(*)
	    into is_any 
	    from user_objects
	    where object_name = p_tree_struct_table and object_type in ('TABLE','VIEW')
	;	
	if is_any <> 0 then
	    error_descr := error_descr ||'Table '||p_tree_struct_table|| ' already exists!';
	    trouble := true;
	end if;
	if trouble then
	   return 0;
	else 
	   return 1;
	end if;
    END check_before_create;
	
    FUNCTION new ( 
        type_id         in generic_tree_types.type_id%TYPE	    default null, 
	owner_id            in generic_tree_types.owner_id%TYPE         default null, 
	type_name	    in generic_tree_types.type_name%TYPE, 
	pretty_name	    in generic_tree_types.pretty_name%TYPE, 
	pretty_plural	    in generic_tree_types.pretty_plural%TYPE, 
	tree_struct_table   in generic_tree_types.tree_struct_table%TYPE, 
	default_function_no in generic_tree_types.default_function_no%TYPE default null,
	creation_date       in acs_objects.creation_date%TYPE		    default sysdate, 
	creation_user       in acs_objects.creation_user%TYPE		    default null, 
	creation_ip         in acs_objects.creation_ip%TYPE		    default null, 
	context_id          in acs_objects.context_id%TYPE		    default null
    ) RETURN generic_tree_types.type_id%TYPE
    IS
        v_type_id integer; 
    BEGIN 
	v_type_id := acs_object.new ( 
	    object_id     => type_id, 
	    object_type   => 'generic_tree_type', 
	    creation_date => creation_date, 
	    creation_user => creation_user, 
	    creation_ip   => creation_ip, 
	    context_id    => context_id 
	); 
	insert into generic_tree_types 
	    (type_id, type_name, tree_struct_table, default_function_no, pretty_name, pretty_plural)
	    values (v_type_id, type_name, tree_struct_table, default_function_no, pretty_name, pretty_plural)
	;
	return v_type_id;
    END new;

    FUNCTION name (
        object_id           in acs_objects.object_id%TYPE
    ) RETURN varchar2
    IS
	obj_name generic_tree_types.type_name%type;
    BEGIN
	select pretty_name 
	    into obj_name 
	    from generic_tree_types 
	    where type_id = object_id
	;
	return obj_name;	
    END name;


    PROCEDURE delete ( 
	type_id         in acs_objects.object_id%TYPE 
    ) 
    IS
    BEGIN 
        delete from generic_tree_types 
	    where type_id = generic_tree_type.delete.type_id
	; 
	acs_object.delete(type_id); 
    END delete;

    FUNCTION generate_db_objects (
	p_type_id       generic_tree_types.type_id%TYPE,
	error_descr	    in out varchar2
    ) RETURN integer
    IS
        statement1 varchar2(4000);
	p_tree_struct_table generic_tree_types.tree_struct_table%TYPE;	
    BEGIN
	select tree_struct_table 
	    into  p_tree_struct_table  
	    from generic_tree_types 
	    where type_id = p_type_id
	;
        
        statement1 := 
	    'create table '|| p_tree_struct_table || ' ('||
	    'node_id	   integer primary key,'|| 
	    'tree_id	   integer references generic_trees (tree_id),'||
	    'object_id	   integer references acs_objects (object_id),'||
	    'parent_node_id integer references '||rtrim(p_tree_struct_table)||' (node_id),'||
	    'deprecated_p   char(1) default ''f'' not null,'||
	    'left_ind	   integer,'||
	    'right_ind	   integer)'
	;
	execute immediate statement1;
	update generic_tree_types set generated_p = 't' where type_id = p_type_id;
	-- create hidden "super" root
	commit;
	return 1; 
    EXCEPTION 
	WHEN OTHERS THEN
	    error_descr := 'Cannot create table '|| p_tree_struct_table || '! Encountered error:'|| DBMS_UTILITY.FORMAT_ERROR_STACK();
	    return 0;
    END generate_db_objects;


    FUNCTION drop_db_objects (
        p_type_id       generic_tree_types.type_id%TYPE,
	errors_descr in out varchar2
    ) RETURN integer
    IS
	p_tree_struct_table generic_tree_types.tree_struct_table%TYPE;	
    BEGIN
	select tree_struct_table 
	    into  p_tree_struct_table  
	    from generic_tree_types 
	    where type_id = p_type_id
	;
	execute immediate 'drop table '|| p_tree_struct_table || 'cascade constraints';
	update generic_tree_types set generated_p = 't' where type_id = p_type_id;
	commit;
	return 1;
    EXCEPTION
	WHEN OTHERS THEN
	     errors_descr :=  'Encountered error:'|| DBMS_UTILITY.FORMAT_ERROR_STACK();
	     return 0;
    END drop_db_objects; 
END;
/
show errors
