-- packages/generic_trees/sql/tree-package.sql
--
-- @creation-date December 2000
--
-- @author Aram Kananov (akananov@arsdigita.com)
--
-- @cvs-id $Id: tree-package.sql,v 1.1 2001/04/27 10:03:05 arcanan Exp $
--
-- TODO List
-- 1)move node/tree
-- 2)nested set api
-- 3)mutator

create or replace package generic_tree
as 

    ------------------------------------------------------------
    -- GLOBAL TYPES AND STATIC VARIABLES
    ------------------------------------------------------------
    TYPE T_NODE is table of integer;
    mutating_nodes T_NODE;

    FUNCTION new ( 
        tree_id		    in generic_trees.tree_id%TYPE		 default null, 
	type_id		    in generic_tree_types.type_id%TYPE,
	tree_name	    in varchar2,
	description	    in varchar2,
	default_function_no in generic_trees.default_function_no%TYPE    default null,
	creation_date       in acs_objects.creation_date%TYPE		 default sysdate, 
	creation_user       in acs_objects.creation_user%TYPE		 default null, 
	creation_ip         in acs_objects.creation_ip%TYPE		 default null, 
	context_id          in acs_objects.context_id%TYPE		 default null
    ) RETURN generic_trees.tree_id%TYPE
    ;

    FUNCTION name (
        object_id           in acs_objects.object_id%TYPE
    ) return varchar2;

    PROCEDURE delete ( 
	tree_id         in acs_objects.object_id%TYPE 
    );

    FUNCTION add_node (
	p_node_id integer default null,
	p_tree_id in generic_trees.tree_id%TYPE, 
	p_object_id  in acs_objects.object_id%TYPE, 
	p_parent_node_id integer
    )	RETURN integer
    ;

    PROCEDURE del_node (
	p_node_id in integer,
	p_tree_id in generic_trees.tree_id%TYPE, 
	p_delete_childs in boolean default false 
    )	
    ;

    PROCEDURE copy_tree (
	      source_tree integer, dest_tree integer, source_node_id integer, dest_parent_node_id integer, depth integer default null
    );

    PROCEDURE move_tree (source_tree integer, dest_tree integer, source_node_id integer, dest_parent_node_id integer);

    -- when the label(value returned by functional index) should be changed we have 
    -- to call before_change proc before update and after_change after update.
    -- otherwise index will become invalid
    PROCEDURE before_change(p_tree_id in generic_trees.tree_id%TYPE, p_object_id  in acs_objects.object_id%TYPE);
    PROCEDURE after_change (p_tree_id in generic_trees.tree_id%TYPE, p_object_id  in acs_objects.object_id%TYPE);

    PROCEDURE refresh_nested_ind (p_tree_id integer);
end; 
/ 
show errors

create or replace package body generic_tree 
as 
    ------------------------------------------------------------
    --	LOCAL FUNCTIONS and PROCEDURES
    ------------------------------------------------------------
    FUNCTION get_struct_table (p_tree_id integer) RETURN varchar2;
    FUNCTION get_table_name (p_type_id integer) RETURN varchar2;
    FUNCTION index_childs(p_table varchar2, p_parent_node_id integer, ind integer) RETURN integer;

    ------------------------------------------------------------
    --	PUBLIC FUNCTIONS and PROCEDURES
    ------------------------------------------------------------
    FUNCTION new ( 
        tree_id		    in generic_trees.tree_id%TYPE	    default null, 
	type_id		    in generic_tree_types.type_id%TYPE,
	tree_name	    in varchar2,
	description	    in varchar2,
	default_function_no in generic_trees.default_function_no%TYPE default null,
	creation_date       in acs_objects.creation_date%TYPE		    default sysdate, 
	creation_user       in acs_objects.creation_user%TYPE		    default null, 
	creation_ip         in acs_objects.creation_ip%TYPE		    default null, 
	context_id          in acs_objects.context_id%TYPE		    default null
    ) RETURN generic_trees.tree_id%TYPE
    IS
        v_tree_id integer;
	ins_root  varchar2(4000); 
	v_node_id integer;
	def_f_no integer;
    BEGIN 
	v_tree_id := acs_object.new ( 
	    object_id     => tree_id, 
	    object_type   => 'generic_tree', 
	    creation_date => creation_date, 
	    creation_user => creation_user, 
	    creation_ip   => creation_ip, 
	    context_id    => context_id 
	); 
	if default_function_no is null then
	    select a.default_function_no 
	        into def_f_no
		from generic_tree_types a 
		where  type_id = generic_tree.new.type_id
	    ;
	else
	    def_f_no := default_function_no;
	end if;
	insert into generic_trees
	    (tree_id, type_id, tree_name, description, default_function_no)
	    values (v_tree_id, type_id, tree_name, description, def_f_no)
	;
	select acs_object_id_seq.nextval into v_node_id from dual;
  	execute immediate 'insert into '||generic_tree.get_table_name (type_id) || 
	    ' (node_id, tree_id) values (:1, :2)' 
	    using v_node_id, v_tree_id
	;
	update generic_trees set root_node_id = v_node_id where tree_id = v_tree_id;
	return v_tree_id;
    END new;

    FUNCTION name (
        object_id           in acs_objects.object_id%TYPE
    ) return varchar2
    IS
        ret generic_trees.tree_name%TYPE;
    BEGIN
        select tree_name into ret from generic_trees where tree_id = object_id;
	return ret;
    END;

    PROCEDURE delete ( 
	tree_id         in acs_objects.object_id%TYPE 
    ) 
    IS
    BEGIN 
        delete from generic_trees
	    where tree_id = generic_tree.delete.tree_id
	; 
	acs_object.delete(tree_id); 
    END delete;

    FUNCTION add_node (
	p_node_id integer default null,
	p_tree_id in generic_trees.tree_id%TYPE, 
	p_object_id  in acs_objects.object_id%TYPE, 
	p_parent_node_id integer
    )	RETURN integer
    IS
	v_node_id integer;
	nested_flag char(1);
	table_name varchar2(50) := get_struct_table (p_tree_id);
	plsql_block varchar2(32000) ;
	v_left_ind integer;
	v_right_ind integer;
    BEGIN
	if p_node_id is null then
	   select acs_object_id_seq.nextval into v_node_id from dual;
	else
	   v_node_id := p_node_id;
	end if; 
 	execute immediate 'insert into '|| table_name || 
                             ' (node_id, tree_id, object_id, parent_node_id) values (:1,:2,:3,:4)' 
	    using v_node_id, p_tree_id, p_object_id, p_parent_node_id
	;

        -- check if nestd set is on and refresh left_ind and right ind columns
	select support_nested_set_p into nested_flag from generic_trees where tree_id = p_tree_id;

	if nested_flag = 't' then

	   plsql_block := 'DECLARE
			       v_left_ind integer;
			       v_right_ind integer;
			    BEGIN 
			       select left_ind, right_ind into v_left_ind, v_right_ind 
			           from  '|| table_name || 
				  ' where node_id = :p_parent_node_id;   
			       update '|| table_name || 
			          ' set left_ind = left_ind + 2,
				    right_ind = right_ind + 2
				    where  tree_id = :p_tree_id 
				       and left_ind > v_left_ind 
                               ;
			       update '|| table_name || 
			          ' set right_ind = right_ind + 2
				    where tree_id = :p_tree_id
				       and left_ind <= v_left_ind 
				       and right_ind >= v_right_ind
                               ;			       
			       update '|| table_name || 
			          ' set left_ind = v_left_ind + 1,
				    right_ind = v_left_ind + 2
				    where node_id = :v_node_id;
			   END;'
           ;

           execute immediate plsql_block using p_parent_node_id, p_tree_id, v_node_id;

	end if;
	return v_node_id;
    END;
    
    PROCEDURE del_node (
	p_node_id in integer,
	p_tree_id in generic_trees.tree_id%TYPE, 
	p_delete_childs in boolean default false 
    )	
    IS
	childs integer;
	sql_text varchar2(4000);
	v_parent_node_id integer;
	nested_flag char(1);
	plsql_block varchar2(32000);
	table_name  varchar2(50) := get_struct_table (p_tree_id);
    BEGIN
	if not p_delete_childs then
	   execute immediate 'select decode(count(*),0,0,1) from '||get_struct_table (p_tree_id) ||' where parent_node_id= :2'
	       into childs using p_node_id;

	   if childs <> 0 then
	       raise_application_error(-20101, 'Delete failed! Node has childs');
	   end if;
 	end if;
        
        execute immediate ' select parent_node_id from '|| table_name || ' where node_id = :p_node_id'
	     into  v_parent_node_id using p_node_id
	; 	

	sql_text := 'delete '|| get_struct_table (p_tree_id) 
		   ||' where node_id in (select node_id from '||get_struct_table (p_tree_id)
                                       ||' start with node_id = :1 connect by prior node_id = parent_node_id)'
	; 
	execute immediate sql_text using p_node_id;
	
        -- check if nestd set is on and refresh left_ind and right ind columns
	select support_nested_set_p into nested_flag from generic_trees where tree_id = p_tree_id;

	if nested_flag = 't' then

	    -- if  deleting subtree then call refresh_nested_id
	    if p_delete_childs then 
	       refresh_nested_ind(p_tree_id);

	    -- if deleteing one node    
	    else
  	        plsql_block := 'DECLARE
  			            v_left_ind integer;
				    v_right_ind integer;
			        BEGIN 
				    select left_ind, right_ind into v_left_ind, v_right_ind 
				        from  '|| table_name || 
					' where node_id = :v_parent_node_id;   
				    update '|| table_name || 
				        ' set left_ind = left_ind -2, right_ind = right_ind - 2 
					where left_ind > v_left_ind and tree_id = :p_tree_id;
			            update '|| table_name || 
				        ' set right_ind = right_ind -2 
					where left_ind <= v_left_ind and tree_id = :p_tree_id;
			        END;'
                ;
	        execute immediate plsql_block using v_parent_node_id, p_tree_id;

	   end if;

       end if;

    END;

    PROCEDURE copy_tree (source_tree integer, 
			 dest_tree integer, 
			 source_node_id integer, 
			 dest_parent_node_id integer, 
			 depth integer default null)
    IS 
        sql_text varchar2(4000);
	nested_flag char(1);
    BEGIN
	sql_text := 'insert into '|| get_struct_table (dest_tree)||
                    '   (node_id, tree_id, object_id, parent_node_id )
		        select acs_object_id_seq.nextval, :dest_tree, object_id, 
                               nvl(prior acs_object_id_seq.nextval,:dest_parent_node_id) 
		               from '|| get_struct_table (source_tree)|| ' 
                               start with node_id = :source_node_id 
                                 connect by prior node_id = parent_node_id '
        ;
	if depth is not null then
	    sql_text := sql_text || ' and level <= :depth';
	    execute immediate sql_text using   dest_tree, dest_parent_node_id, source_node_id, depth;
	else 
	    execute immediate sql_text using   dest_tree, dest_parent_node_id, source_node_id;
	end if;

        -- check if nested set is on and refresh left_ind and right ind columns
	select support_nested_set_p into nested_flag from generic_trees where tree_id = dest_tree;

	if nested_flag = 't' then
	    refresh_nested_ind(dest_tree);
	end if;	   

    END;

    PROCEDURE move_tree (source_tree integer, dest_tree integer, source_node_id integer, dest_parent_node_id integer)
    IS
	source_table varchar2(50) :=  get_struct_table(dest_tree);
	dest_table   varchar2(50) :=  get_struct_table(source_tree);
	sql_text     varchar2(4000);
	nested_flag char(1);
    BEGIN
	-- trees are in the same table
	if source_table = dest_table then
	    sql_text := 'update '||source_table||' set tree_id = :dest_tree where node_id in (select node_id from '
		     ||source_table|| ' start with node_id = :source_node_id connect by prior node_id = parent_node_id)'
            ;
            execute immediate sql_text using  dest_tree, source_node_id;
	    sql_text := 'update ' ||source_table||' set parent_node_id = :dest_parent_node_id where node_id = :source_node_id';
	    execute immediate sql_text using dest_parent_node_id, source_node_id;
	else
	    sql_text := 'insert into '||dest_table||' (node_id, tree_id, object_id, parent_node_id, deprecated_p)' ||
			'  select node_id, tree_id, object_id, parent_node_id, deprecated_p from '||source_table||
			       ' start with node_id = :source_node_id connect by prior node_id = parent_node_id'  
	    ;
	    execute immediate sql_text using source_node_id;
	    sql_text := 'update '||dest_table||'  set parent_node_id = :dest_parent_node_id where node_id = :source_node_id';
	    execute immediate sql_text using dest_parent_node_id;
	    sql_text := 'delete from '||source_table||' where node_id in (select node_id from ' ||source_table|| 
				     ' start with node_id = :source_node_id connect by prior node_id = parent_node_id)'
	    ;
	    execute immediate sql_text using source_node_id;				  
	end if;

        -- check if nested set is on and refresh left_ind and right ind columns
	select support_nested_set_p into nested_flag from generic_trees where tree_id = dest_tree;

	if nested_flag = 't' then
	    refresh_nested_ind(dest_tree);
	end if;	

    END;


    -- this proc populates mutating_nodes array with node_id's which belog to given object_id 
    -- then sets object_id to null in apropritate rows to prevent index invalidating 
    PROCEDURE before_change (p_tree_id in generic_trees.tree_id%TYPE, p_object_id  in acs_objects.object_id%TYPE)
    IS
        plsql_block varchar2(32000);
	table_name  varchar2(50) := get_struct_table (p_tree_id);
    BEGIN
        -- reset array
        mutating_nodes := null;
        plsql_block := 'BEGIN
		            select node_id
		                bulk collect into generic_tree.mutating_nodes
			        from ' || table_name || '
			        where object_id = :p_object_id
			    ;
			    update ' || table_name || ' set object_id = null where object_id = :p_object_id ;
			END;' 			
        ;
	execute immediate plsql_block using p_object_id;
    END before_change;

    PROCEDURE after_change (p_tree_id in generic_trees.tree_id%TYPE, p_object_id  in acs_objects.object_id%TYPE)
    IS 
        plsql_block varchar2(32000);
	table_name  varchar2(50) := get_struct_table (p_tree_id);
    BEGIN
        plsql_block := 'begin forall i in 1..generic_tree.mutating_nodes.last
		          update ' || table_name || ' set object_id = :p_object_id where node_id = generic_tree.mutating_nodes(i); end;'
        ;
        execute immediate plsql_block using p_object_id; 
    END after_change;

    PROCEDURE refresh_nested_ind (p_tree_id integer)
    IS
       p_table varchar2(50) := get_struct_table (p_tree_id);
       p_root_node_id integer;
       v_ind integer;
    BEGIN
       select root_node_id into p_root_node_id from generic_trees where tree_id = p_tree_id;
       v_ind := index_childs(p_table, p_root_node_id, 1);
       execute immediate 'update '|| p_table || ' set left_ind = :1, right_ind = :2 where node_id = :3' 
           using 1, v_ind + 1,  p_root_node_id 
       ; 	
    END;        


    ---------------------------------------------------------------------

    FUNCTION get_struct_table (p_tree_id integer) RETURN varchar2
    IS
        table_name varchar2(50);
    BEGIN
	select tree_struct_table
	    into table_name 
	    from generic_trees a, generic_tree_types b
	    where a.tree_id = p_tree_id 
	        and a.type_id = b.type_id
	;
	return table_name;
    END;

    FUNCTION get_table_name (p_type_id integer) RETURN varchar2
    IS
        table_name varchar2(50);
    BEGIN
	select tree_struct_table
	    into table_name 
	    from generic_tree_types b
	    where type_id = p_type_id
	;
	return table_name;
    END;

    FUNCTION index_childs(p_table varchar2, p_parent_node_id integer, ind integer) return integer 
    IS
	TYPE type_nodes IS TABLE OF integer;
	nodes type_nodes := type_nodes();
	TYPE type_cursor IS REF CURSOR ;
	cc type_cursor;
	i integer := 0;
	v_node_id integer;
	v_ind integer := ind;
	l_ind integer;
	r_ind integer;
    BEGIN
	open cc for 'select node_id from ' || p_table || ' where parent_node_id = :p_parent_node_id ' using p_parent_node_id;
	loop	    
	    fetch cc into v_node_id;
	    exit when cc%NOTFOUND;
	    i := i + 1; 
	    nodes.extend;
	    nodes(i) := v_node_id;
	end loop;
	close cc;
	if i > 0 then
	   for j in 1..i loop
	        v_ind := v_ind + 1;
	        l_ind := v_ind;
	        v_ind:= index_childs(p_table, nodes(j), v_ind);
	        v_ind := v_ind + 1;
		execute immediate 'update '|| p_table || ' set left_ind = :1, right_ind = :2 where node_id = :3' 
		   using l_ind, v_ind, nodes(j)
		;	
	   end loop;
	end if;
	return v_ind;
    END;


end;
/
show errors















